"""
see license https://github.com/DerwenAI/kglab#license-and-copyright
"""
from .kglab import KnowledgeGraph

from .graph import NodeRef, PropertyStore

from .subg import Subgraph, SubgraphMatrix, SubgraphTensor

from .topo import Measure, Simplex0, Simplex1

from .pkg_types import RDF_Triple, RDF_Node, SPARQL_Bindings, \
    PathLike, IOPathLike, \
    NodeLike, GraphLike, \
    Census_Item, Census_Dyad_Tally

from .external_import import import_from_neo4j

from .gpviz import GPViz

from .util import get_gpu_count, calc_quantile_bins, stripe_column, root_mean_square

from .version import MIN_PY_VERSION, _versify, _check_version, __version__


# deprecated:
#from .esp import ShapeFactory, EvoShape, EvoShapeNode, EvoShapeEdge, \
#    Leaderboard, SerializedEvoShape, EvoShapeBoard, EvoShapeDistance

#from .srl import PSLModel
